/* gts2pov.c
 * convert GNU Triangulated Surfaces (GTS) data to PovRay mesh2.
 */

/* setreuid(2) */
#define _XOPEN_SOURCE 700
/* strdup(3) */
#define _POSIX_C_SOURCE 200809L

#include <assert.h>
#include <errno.h>
#include <locale.h>
#include <regex.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

/* conveniences + misc constants */
#define DFLTIDENT    "mesh"
#define DFLTPREFIX   "gts_"
#define DFLTVERSION  3.8

#define CHRNUL  '\0'

#define EEXIT  (eExit(__func__,__LINE__))

/* longest identifier and prefix parts */
#define IDENTMAX  40

/* only need to limit vertices, configurable */
#ifndef GTSVMAX
#define GTSVMAX  1000000
#endif

/* all lines max 2046 chars excl, oversized,
 * no idea how much extra tools can append to data.
 */
#define LNBUFSZ  2048

/* want unsigned integers and floats from gts data.  numbers.. :-( */
#define RE_N  "[-+]?([[:digit:]]*[.]?[[:digit:]]+|[[:digit:]]+[.][[:digit:]]*)"
#define RE_E  "([Ee][-+]?[[:digit:]]+)?"
#define REFLT "(" RE_N RE_E ")"
#define REINT  "([[:digit:]]+)"
#define RESPC  "[ \t]"

/* as 'gts.h', plus semicolon.  first column only */
#define GTS_COMMENT  "!;#"

/* custom types.  a three element vector for vertices,
 * two and three (index) element structs for edges and
 * faces, and inside_vector axis.
 */
enum InsideVect { kNone = 0, kX, kY, kZ };
typedef enum InsideVect InsideVect;

typedef struct Vertex Vertex;
struct Vertex { double x, y, z; };

typedef struct Edge Edge;
struct Edge { size_t i1, i2; };

typedef struct Face Face;
struct Face { size_t i1, i2, i3; };

/* file global.  CLI stuff, mostly */
static char const * const prgname = "gts2pov";
static char const * const prgversion = "0.2.0";
/* -1       mesh.            default 'mesh2'.
   -b       "bare", no include wrapper.
   -a file  auxiliary.       default none.
   -h       help.
   -i axis  inside vector.   default none.
   -n name  identifier.      default 'mesh'.
   -P n.n   povray version.  default '3.8'.
   -p abc   common prefix.   default 'gts_'.
   -[Vv]    version.
 */
static char const * const prgopts = "+1a:bhi:n:P:p:Vv";
/* only one message for '-[Vv]' */
static char const * const prgopts_help[] = {
  "1\t\tgenerate older 'mesh' type instead",
  "a filename\t'[object modifiers]' to insert",
  "b\t\tgenerate \"bare\" object, no .inc wrap",
  "h\t\tdisplay this message",
  "i x|y|z\tinside vector for mesh",
  "n name\tmesh name/identifier",
  "P n.n\tactive POV-Ray version",
  "p name\tcommon identifier prefix",
  "V\t\tdisplay the program version",
  NULL
};

static char *aux_value;
static FILE *aux_fp;
static bool mkbare;
static bool mkmesh2;
static double povvers_value;
static char *prefix_value;
static char *ident_value;
static InsideVect inside;

/* the GTS spec allows any data line to be extended with additional
 * data, hence no end anchors.
 * NB. shared format counts and face.
 */
enum REIndex {
  kComment = 0, kEmpty, kVertex, kEdge, kFace, kCounts = 4, kendRE
};
typedef enum REIndex REIndex;

static char const * const re_patterns [kendRE] = {
  "^[" GTS_COMMENT "]",
  "^\n$",
  "^" RESPC "*" REFLT RESPC "+" REFLT RESPC "+" REFLT,
  "^" RESPC "*" REINT RESPC "+" REINT,
  "^" RESPC "*" REINT RESPC "+" REINT RESPC "+" REINT
};
static regex_t rgx[kendRE];

// TODO RE for name + prefix?

static char const * const inside_vectors [] = {
  "", "<1,0,0>", "<0,1,0>", "<0,0,1>", NULL
};

/* vertex, edge, and face counts on first data line */
static size_t nV, nE, nF;
/* arrays for {nV} + {nE} elements plus one, cf 1-based */
static Vertex *arrV;
static Edge   *arrE;

static char const * const eofmask = "%s: oops, unexpected EOF on <stdin>.\n",
                  * const ivecmask = "  inside_vector %s\n",
                  * const auxmask =
        "// ----------<auxiliary %s '%s'>----------\n";


/* -------------------------------------------------------------------------- *
 * oops + GNU mandated interface.
 */

static void eExit(char const * const, int const) __attribute__((noreturn));
static void eExit(char const * const fn, int const ln) {
  int const e = errno;
  fprintf(stderr, "%s: oops in '%s' line %d, error (%d) %s.\n",
          prgname, fn, ln, e, strerror(e));
  exit(EXIT_FAILURE);
}

static void usageHelp(int const) __attribute__((noreturn));
static void usageHelp(int const ec) {
  char const * const fmt = "%s - GTS data to POV-Ray 'mesh2'.\n"
          "\n  usage: %s [options]\n"
          "\navailable options are:\n";
  printf(fmt, prgname, prgname);
  for (size_t i = 0; prgopts_help[i]; ++i)
      printf("  -%s.\n", prgopts_help[i]);
  printf("\n");
  exit(ec);
}

static void usageVersion(void) __attribute__((noreturn));
static void usageVersion(void) {
  char const * const fmt = "%s %s\n"
          "copyright (c) 2020, jr <creature.eternal@gmail.com>.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to .. whatever.\n"
          "there is no warranty.\n";
  printf(fmt, prgname, prgversion);
  exit(EXIT_SUCCESS);
}

/* -------------------------------------------------------------------------- *
 * helpers.
 */

static bool getDouble(char const * const arg, double *val) {
  errno = 0;
  char *p = NULL;
  double d = strtod(arg, &p);
  if (errno || *p) return false;
  *val = d;
  return true;
}

/* return 'true' and next data line of given type, and 'false' when EOF.
 * exits on error.
 */
static bool getDataLine(char * const line, size_t const i) {
  static unsigned long int lnum = 0;

  assert(line  &&  !(*line));

  char buf[LNBUFSZ];

label_get_another:
  memset(buf, 0, LNBUFSZ);
  ++lnum;
  if (!fgets(buf, LNBUFSZ, stdin)) {
    if (feof(stdin)) return false;
    EEXIT;
  }
  /* skip comments and empty lines, verify shape as expected */
  if (!regexec(&rgx[kComment], buf, 0, NULL, 0))
    goto label_get_another;
  else if (!regexec(&rgx[kEmpty], buf, 0, NULL, 0))
    goto label_get_another;
  else if (regexec(&rgx[i], buf, 0, NULL, 0)) {
    fprintf(stderr, "%s: oops, bad data in line %lu.\n", prgname, lnum);
    exit(EXIT_FAILURE);
  }

  strcpy(line, buf);
  return true;
}

static void edgeFromLine(char const * const line, Edge *data) {
  char const * const fmt = " %zu %zu";
  if (regexec(&rgx[kEdge], line, 0, NULL, 0)) EEXIT;
  size_t a, b;  a = b = 0;
  if (2 != sscanf(line, fmt, &a, &b)) EEXIT;
  *data = (Edge){a, b};
  return;
}

static void faceFromLine(char const * const line, Face *data) {
  char const * const fmt = " %zu %zu %zu";
  if (regexec(&rgx[kFace], line, 0, NULL, 0)) EEXIT;
  size_t a, b, c;  a = b = c = 0;
  if (3 != sscanf(line, fmt, &a, &b, &c)) EEXIT;
  *data = (Face){a, b, c};
  return;
}

static void vertexFromLine(char const * const line, Vertex *data) {
  char const * const fmt = " %lg %lg %lg";
  if (regexec(&rgx[kVertex], line, 0, NULL, 0)) EEXIT;
  double x, y, z;  x = y = z = 0;
  if (3 != sscanf(line, fmt, &x, &y, &z)) EEXIT;
  *data = (Vertex){x, y, z};
  return;
}

static Face e2v(Face const a) {
  size_t v1 = arrE[a.i1].i1;
  size_t v2 = arrE[a.i1].i2;
  size_t v3;
  if (arrE[a.i2].i1 == v1  ||  arrE[a.i2].i1 == v2)
    v3 = arrE[a.i2].i2;
  else
    v3 = arrE[a.i2].i1;
  return (Face){v1, v2, v3};
}

/* output face indices are zero-based */
static Face zidx(Face const a) {return (Face){a.i1 - 1, a.i2 - 1, a.i3 - 1};}

/* -------------------------------------------------------------------------- *
 * output.
 */

static void emitPreamble(void) {
  char const * const fmt = "// GTS surface/mesh data.\n"
          "// converted %s (%s v %s).\n\n"
          "#ifndef(%sinclude_temp)\n\n"
          "#declare %sinclude_temp = version;\n"
          "#version %g;\n\n"
          "#ifdef(View_POV_Include_Stack)\n"
          "#debug \"including %s%s.inc\\n\"\n"
          "#end\n\n";
  time_t const now = time(0);
  struct tm *tm = localtime(&now);
  char buf[41] = {CHRNUL};
  strftime(buf, 40, "%d.%m.%Y %H%Mh", tm);

  if (mkbare)
    printf("// converted %s (%s v %s).\n", buf, prgname, prgversion);
  else
    printf(fmt, buf, prgname, prgversion, prefix_value, prefix_value,
            povvers_value, prefix_value, ident_value);
  return;
}

static void emitPostamble(void) {
  char const * const fmt = "\n#version %sinclude_temp;\n\n#end\n\n";
  if (!mkbare)
    printf(fmt, prefix_value);
  return;
}

static void emitAuxiliary(void) {
  printf(auxmask, "begin", aux_value);
  char buf[LNBUFSZ] = {CHRNUL};
  while (fgets(buf, LNBUFSZ, aux_fp))
    if (EOF == fputs(buf, stdout)) EEXIT;
  if (ferror(aux_fp)) EEXIT;
  printf(auxmask, "end", aux_value);
  return;
}

static void emitMesh(void) {
  printf("#declare %s%s = mesh {\n", prefix_value, ident_value);
  char buf[LNBUFSZ];
  for (size_t i = 0; nF != i; ++i) {
    memset(buf, 0, LNBUFSZ);
    if (!getDataLine(buf, kFace)) {
      fprintf(stderr, eofmask, prgname);
      exit(EXIT_FAILURE);
    }
    Face a = {0};
    faceFromLine(buf,&a);
    assert((a.i1 * a.i2 * a.i3) && nE >= a.i1 && nE >= a.i2 && nE >= a.i3);
    /* swap edge indices for vertex indices */
    a = e2v(a);
    printf("  triangle {<%g,%g,%g>, <%g,%g,%g>, <%g,%g,%g>}\n",
            arrV[a.i1].x, arrV[a.i1].y, arrV[a.i1].z,
            arrV[a.i2].x, arrV[a.i2].y, arrV[a.i2].z,
            arrV[a.i3].x, arrV[a.i3].y, arrV[a.i3].z);
  }
  if (inside) printf(ivecmask, inside_vectors[inside]);
  if (aux_value) emitAuxiliary();
  printf("};\n");
  return;
}

static void emitMesh2(void) {
  printf("#declare %s%s = mesh2 {\n", prefix_value, ident_value);
  printf("  vertex_vectors {\n    %zu,\n", nV);
  for (size_t i = 1; nV >= i; ++i)
    printf("    <%g,%g,%g>%s\n",
            arrV[i].x, arrV[i].y, arrV[i].z, (nV < (i + 1)) ? "" : ",");
  printf("  }\n  face_indices {\n    %zu,\n", nF);
  char buf[LNBUFSZ];
  for (size_t i = 0; nF != i; ++i) {
    memset(buf, 0, LNBUFSZ);
    if (!getDataLine(buf, kFace)) {
      fprintf(stderr, eofmask, prgname);
      exit(EXIT_FAILURE);
    }
    Face a = {0};
    faceFromLine(buf,&a);
    assert((a.i1 * a.i2 * a.i3) && nE >= a.i1 && nE >= a.i2 && nE >= a.i3);
    /* swap indices, plus adjust to 0-based index */
    a = zidx(e2v(a));
    printf("    <%zu,%zu,%zu>%s\n",
            a.i1, a.i2, a.i3, (nF == i + 1)  ? "" : ",");
  }
  printf("  }\n");
  if (inside) printf(ivecmask, inside_vectors[inside]);
  if (aux_value) emitAuxiliary();
  printf("};\n");
  return;
}

/* -------------------------------------------------------------------------- *
 * 'convert' controls sequence.
 */

static void getCounts(void) {
  char buf[LNBUFSZ] = {CHRNUL};
  if (!getDataLine(buf, kCounts)) {
    fprintf(stderr, eofmask, prgname);
    exit(EXIT_FAILURE);
  }
  /* "misuse" the (identical format) Face type */
  Face tmp = {0};
  faceFromLine(buf, &tmp);
  nV = tmp.i1;
  nE = tmp.i2;
  nF = tmp.i3;
  assert(nV * nE * nF);
  assert(GTSVMAX >= nV);
  return;
}

static void getVertices(void) {
  arrV = calloc(nV + 1, sizeof(Vertex));
  if (!arrV) EEXIT;
  for (size_t i = 1; nV >= i; ++i) {
    char buf[LNBUFSZ] = {CHRNUL};
    if (!getDataLine(buf, kVertex)) {
      fprintf(stderr, eofmask, prgname);
      exit(EXIT_FAILURE);
    }
    vertexFromLine(buf, arrV + i);
  }
  return;
}

static void getEdges(void) {
  arrE = calloc(nE + 1, sizeof(Edge));
  if (!arrE) EEXIT;
  for (size_t i = 1; nE >= i; ++i) {
    char buf[LNBUFSZ] = {CHRNUL};
    if (!getDataLine(buf, kEdge)) {
      fprintf(stderr, eofmask, prgname);
      exit(EXIT_FAILURE);
    }
    Edge tmp = {0};
    edgeFromLine(buf, &tmp);
    assert((tmp.i1 * tmp.i2) && nV >= tmp.i1 && nV >= tmp.i2);
    arrE[i] = tmp;
  }
  return;
}

static int convert(void) {
  getCounts();
  getVertices();
  getEdges();
  emitPreamble();
  if (mkmesh2) emitMesh2();
  else emitMesh();
  emitPostamble();
  return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- *
 * specific function ('argXxx') for each option with an argument,
 * true when successful.
 */

/* NB. not bool */
static void argAuxiliary(char const * const arg) {
  aux_fp = fopen(arg, "rb");
  if (!aux_fp) {
    fprintf(stderr, "%s: oops, opening '%s' failed.\n", prgname, arg);
    exit(1);
  }
  aux_value = strdup(arg);
  if (!aux_value) {
    (void)fclose(aux_fp);
    EEXIT;
  }
  return;
}

static bool argInside(char const * const arg) {
  bool rf = true;
  switch (*arg) {
    case 'X': case 'x': inside = kX; break;
    case 'Y': case 'y': inside = kY; break;
    case 'Z': case 'z': inside = kZ; break;
    default:            rf = false;
  }
  return rf;
}

// TODO verify content.
static bool argName(char const * const arg) {
  size_t const n = strlen(arg);
  if (!n || (IDENTMAX <= n))
    return false;
  ident_value = strdup(arg);
  if (!ident_value) EEXIT;
  return true;
}

/* NB. PovRay version magic numbers */
static bool argPovvers(char const * const arg) {
  double d = 0;
  if (!getDouble(arg, &d))
    return false;
  else if (3.5 > d  ||  4.0 <= d)
    return false;
  povvers_value = d;
  return true;
}

// TODO as 'argName'.
static bool argPrefix(char const * const arg) {
  size_t const n = strlen(arg);
  if (!n || (IDENTMAX <= n))
    return false;
  prefix_value = strdup(arg);
  if (!prefix_value) EEXIT;
  return true;
}

/* -------------------------------------------------------------------------- *
 * main's job, initialisation aside, is dealing with the arguments.
 * returns via the convert function.
 */

static void tidyup(void) {
  if (aux_value) (void)fclose(aux_fp);
  free(arrV);
  free(arrE);
  free(aux_value);
  free(ident_value);
  free(prefix_value);
  for (size_t i = 0; kendRE != i; ++i)
    regfree(&rgx[i]);
  return;
}

/* "sensible" defaults, zeroing redundant, but hey.. */
static void setDefaults(void) {
  for (size_t i = 0; kendRE != i; ++i)
    if (regcomp(&rgx[i], re_patterns[i], REG_EXTENDED|REG_NOSUB)) EEXIT;
  aux_value = NULL;
  ident_value = NULL;
  inside = kNone;
  mkbare = false;
  mkmesh2 = true;
  povvers_value = DFLTVERSION;
  prefix_value = NULL;
  return;
}

int main(int argc, char *argv[]) {
  if (atexit(tidyup)) EEXIT;
  else if (setreuid(getuid(), geteuid())) EEXIT;
  else if (!setlocale(LC_ALL, "")) EEXIT;
  setDefaults();
  bool have1, havebare, haveins, havename, havepref, havevers;
  int o;
  have1 = havebare = haveins = havename = havepref = havevers = false;
  opterr = 0;
  while (-1 != (o = getopt(argc, argv, prgopts)))
    switch (o) {
      case '1':
        if (have1) {
          fprintf(stderr, "%s: oops, too many 'mesh' arguments.\n", prgname);
          exit(EXIT_FAILURE);
        }
        mkmesh2 = false;
        have1 = true;
        break;
      case 'a':
        if (aux_value) {
          fprintf(stderr, "%s: oops, too many auxiliary files.\n", prgname);
          exit(EXIT_FAILURE);
        }
        argAuxiliary(optarg);
        break;
      case 'b':
        if (havebare) {
          fprintf(stderr, "%s: oops, too many \"bare\" arguments.\n", prgname);
          exit(EXIT_FAILURE);
        }
        mkbare = true;
        havebare = true;
        break;
      case 'h':  usageHelp(0);
      case 'i':
        if (haveins || !argInside(optarg)) {
          fprintf(stderr, "%s: oops, bad \"inside\" argument.\n", prgname);
          exit(EXIT_FAILURE);
        }
        haveins = true;
        break;
      case 'n':
        if (havename || !argName(optarg)) {
          fprintf(stderr, "%s: oops, bad identifier.\n", prgname);
          exit(EXIT_FAILURE);
        }
        havename = true;
        break;
      case 'P':
        if (havevers || !argPovvers(optarg)) {
          fprintf(stderr, "%s: oops, need one \"version\", 3.5 to 3.8.\n",
                  prgname);
          exit(EXIT_FAILURE);
        }
        havevers = true;
        break;
      case 'p':
        if (havepref || !argPrefix(optarg)) {
          fprintf(stderr, "%s: oops, bad identifier prefix.\n", prgname);
          exit(EXIT_FAILURE);
        }
        havepref = true;
        break;
      case 'V': case 'v':  usageVersion();
      default:   usageHelp(2);
    }

  if (argc != optind) {
    fprintf(stderr, "%s: oops, extra argument(s) on command line.\n",
            prgname);
    exit(EXIT_FAILURE);
  }
  /* final preps */
  if (!havename) if (!(ident_value = strdup(DFLTIDENT))) EEXIT;
  if (!havepref) if (!(prefix_value = strdup(DFLTPREFIX))) EEXIT;

  return convert();
}

/* ------------------------------------------------------------- *
 * the content above is covered by the GPLv3+, see file COPYING. *
 * copyright (c) 2020 jr <creature.eternal@gmail.com>.           *
 * all rights reserved.                                          *
 * ------------------------------------------------------------- *
 */
